
Require Import Setoid Morphisms.
Require Import Omega.

Module Type Eqv.
  Parameter Inline t : Type.
  Parameter Inline eq : t -> t -> Prop.
  Parameter eq_equiv : Equivalence eq.
  Existing Instance eq_equiv.
End Eqv.

Module Make (X:Eqv).
  Import X.

  Definition map := nat -> t.
  Definition eq_map : relation map := (pointwise_relation nat eq).

  Definition nil (x:t) : map := fun _ => x.
Definition cons (x:t) (i:map) : map :=
  fun k => match k with
  | 0 => x
  | S n => i n
  end.
Definition shift (n:nat) (i:map) : map := fun k => i (n+k).

Definition lams n (f:map -> map) (i:map) : map :=
  fun k => if le_gt_dec n k then f (shift n i) (k-n) else i k.

Instance cons_morph : Proper (eq ==> eq_map ==> eq_map) cons.
do 5 red; intros.
destruct a; simpl; trivial.
Qed.
Instance cons_morph' :
  Proper (eq ==> eq_map ==> @Logic.eq _ ==> eq) cons.
do 4 red; intros.
subst y1; destruct x1; simpl; trivial.
Qed.

Instance shift_morph : Proper (@Logic.eq _ ==> eq_map ==> eq_map) shift.
do 5 red; intros.
subst y.
unfold shift; apply H0.
Qed.

Instance lams_morph :
  Proper (@Logic.eq _ ==> (eq_map ==> eq_map) ==> eq_map ==> eq_map) lams.
do 6 red; intros.
subst y.
unfold lams.
destruct (le_gt_dec x a); eauto.
apply H0.
rewrite H1.
reflexivity.
Qed.

Lemma cons_ext : forall (x:t) i i',
  eq x (i' 0) ->
  eq_map i (shift 1 i') ->
  eq_map (cons x i) i'.
do 2 red; intros.
destruct a; simpl; auto.
apply H0.
Qed.

Lemma shift0 : forall i, eq_map (shift 0 i) i.
do 2 red; reflexivity.
Qed.

Lemma shift_split : forall n i, 
  eq_map (shift (S n) i) (shift n (shift 1 i)).
do 2 red; unfold shift; intros.
replace (1+(n+a)) with (1+n+a); simpl; auto with *.
Qed.

Lemma shift_cons : forall x i,
  eq_map (shift 1 (cons x i)) i.
do 2 red; intros.
reflexivity.
Qed.

Lemma lams_bv : forall m f i k,
  k < m -> eq (lams m f i k) (i k).
unfold lams; intros.
destruct (le_gt_dec m k).
 apply False_ind; omega.
 reflexivity.
Qed.

Lemma lams_shift : forall m f i,
  eq_map (shift m (lams m f i)) (f (shift m i)).
unfold lams, shift; do 2 red; intros.
destruct (le_gt_dec m (m+a)).
 replace (m+a-m) with a by omega.
 reflexivity.
apply False_ind; omega.
Qed.

Lemma lams0 : forall f i, eq_map (lams 0 f i) (f (fun k => i k)).
intros.
rewrite <- (shift0 (lams 0 f i)).
rewrite lams_shift.
reflexivity.
Qed.


Lemma shift_lams : forall k f i,
  eq_map (shift 1 (lams (S k) f i)) (lams k f (shift 1 i)).
do 2 red; unfold lams, shift; simpl; intros.
destruct (le_gt_dec k a); simpl; reflexivity.
Qed.

Lemma cons_lams :
  forall  k f i x,
  Proper (eq_map ==> eq_map) f ->
  eq_map (cons x (lams k f i)) (lams (S k) f (cons x i)).
intros.
apply cons_ext; simpl.
 rewrite lams_bv; auto with *.
 rewrite shift_lams.
 rewrite shift_cons; reflexivity.
Qed.

End Make.

(**)
(*
Definition intt := nat -> Lc.term.
Definition eq_intt := pointwise_relation nat (@eq Lc.term).
Instance eq_intt_equiv : Equivalence eq_intt.
Definition icons (t:Lc.term) (j:intt) : intt :=
  fun k => match k with
  | 0 => t
  | S n => j n
  end.
Definition ishift (n:nat) (j:intt) : intt :=
  fun k => j (n+k).
Definition ilams n (f:intt -> intt) (j:intt) : intt :=
  fun k => if le_gt_dec n k then f (ishift n j) (k-n) else j k.


Instance icons_morph : Proper (@eq _ ==> eq_intt ==> eq_intt) icons.
do 5 red; intros.
destruct a; simpl; trivial.
Qed.

Instance ishift_morph : Proper (@eq _ ==> eq_intt ==> eq_intt) ishift.
do 5 red; intros.
unfold ishift; subst y; auto.
Qed.

Instance ilams_morph :
  Proper (@eq _ ==> (eq_intt ==> eq_intt) ==> eq_intt ==> eq_intt) ilams.
do 6 red; intros.
subst y.
unfold ilams.
destruct (le_gt_dec x a); eauto.
apply H0.
rewrite H1.
reflexivity.
Qed.

Instance ilift_morph : Proper (eq_intt ==> eq_intt) ilift.
do 4 red; intros.
unfold ilift.
destruct a; simpl; auto.
rewrite H; trivial.
Qed.

Lemma icons_ext : forall x j j',
  x = j' 0 ->
  eq_intt j (ishift 1 j') ->
  eq_intt (icons x j) j'.
do 2 red; intros.
destruct a; simpl; auto.
apply H0.
Qed.

Lemma ishift0 : forall j, eq_intt (ishift 0 j) j.
do 2 red; reflexivity.
Qed.

Lemma ishift_split : forall n j, 
  eq_intt (ishift (S n) j) (ishift n (ishift 1 j)).
do 2 red; unfold shift; intros.
replace (1+(n+a)) with (1+n+a); simpl; auto with *.
Qed.

Lemma ishift_cons : forall x j,
  eq_intt (ishift 1 (icons x j)) j.
do 2 red; intros.
reflexivity.
Qed.

Lemma ilams_bv : forall m f j k,
  k < m -> ilams m f j k = j k.
unfold ilams; intros.
destruct (le_gt_dec m k).
 apply False_ind; omega.
 reflexivity.
Qed.

Lemma ilams_shift : forall m f j,
  eq_intt (ishift m (ilams m f j)) (f (ishift m j)).
unfold ilams, ishift; do 2 red; intros.
destruct (le_gt_dec m (m+a)).
 replace (m+a-m) with a by omega.
 reflexivity.
apply False_ind; omega.
Qed.

Lemma ilams0 : forall f j, eq_intt (ilams 0 f j) (f (fun k => j k)).
intros.
rewrite <- (ishift0 (ilams 0 f j)).
rewrite ilams_shift.
reflexivity.
Qed.

Lemma ishift_lams : forall k f j,
  eq_intt (ishift 1 (ilams (S k) f j)) (ilams k f (ishift 1 j)).
do 2 red; unfold ilams, ishift; simpl; intros.
destruct (le_gt_dec k a); simpl; reflexivity.
Qed.

Lemma ilift_lams : forall k f j,
  (forall j j', (forall a, Lc.lift 1 (j a) = j' a) ->
   forall a, Lc.lift 1 (f j a) = f j' a) ->
  eq_intt (ilift (ilams k f j)) (ilams (S k) f (ilift j)).
intros.
do 2 red; intros.
destruct a; simpl.
 reflexivity.

 unfold ilams; simpl.
 destruct (le_gt_dec k a); simpl; trivial.
 apply H.
 unfold ishift, ilift; simpl; intros; trivial.
Qed.

Lemma ilift_binder : forall u j k,
  eq_intt
    (ilift (fun n => Lc.subst_rec u (j n) k))
    (fun n => Lc.subst_rec u (ilift j n) (S k)).
red; red; intros.
destruct a; simpl; trivial.
rewrite Lc.commut_lift_subst; trivial.
Qed.

Lemma ilift_binder_lift : forall j k,
  eq_intt
    (ilift (fun n => Lc.lift_rec 1 (j n) k))
    (fun n => Lc.lift_rec 1 (ilift j n) (S k)).
red; red; intros.
destruct a; simpl; trivial.
rewrite Lc.permute_lift; trivial.
Qed.
*)